/*
 * Copyright (C) 1997 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file attribute.c Attribute handling functions. */

#include <regex.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "ipc.h"
#include "sideband.h"
#include "command.h"

static struct osl_table *attribute_table;
static int greatest_att_bitnum;

/** The columns of the attribute table. */
enum attribute_table_columns {
	/** The bit number (0-63). */
	ATTCOL_BITNUM,
	/** The name of the attribute. */
	ATTCOL_NAME,
	/** Number of columns in this table. */
	NUM_ATT_COLUMNS
};

static int char_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	const unsigned char *c1 = (const unsigned char*)obj1->data;
	const unsigned char *c2 = (const unsigned char*)obj2->data;
	if (*c1 > *c2)
		return 1;
	if (*c1 < *c2)
		return -1;
	return 0;
}

static struct osl_column_description att_cols[] = {
	[ATTCOL_BITNUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "bitnum",
		.compare_function = char_compare,
		.data_size = 1
	},
	[ATTCOL_NAME] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_UNIQUE,
		.name = "name",
		.compare_function = string_compare,
	}
};

static struct osl_table_description attribute_table_desc = {
	.name = "attributes",
	.num_columns = NUM_ATT_COLUMNS,
	.flags = 0,
	.column_descriptions = att_cols
};

static void find_greatest_att_bitnum(void)
{
	unsigned char c = 63;
	do {
		struct osl_row *row;
		struct osl_object obj = {.data = &c, .size = 1};
		if (osl_get_row(attribute_table, ATTCOL_BITNUM, &obj,
				&row) >= 0) {
			greatest_att_bitnum = c;
			return;
		}
	} while (c--);
	PARA_INFO_LOG("no attributes\n");
	greatest_att_bitnum = -E_NO_ATTRIBUTES;
}

/**
 * Retrieve the identifier (number) of an attribute.
 *
 * \param att_name The name of the attribute.
 * \param bitnum Result pointer.
 *
 * \return Positive on success, negative on errors.
 */
int get_attribute_bitnum_by_name(const char *att_name, unsigned char *bitnum)
{
	struct osl_object obj = {.data = (char *)att_name,
		.size = strlen(att_name) + 1};
	struct osl_row *row;
	int ret = osl(osl_get_row(attribute_table, ATTCOL_NAME, &obj, &row));

	if (ret < 0)
		return ret;
	ret = osl(osl_get_object(attribute_table, row, ATTCOL_BITNUM, &obj));
	if (ret < 0)
		return ret;
	*bitnum = *(unsigned char *)obj.data;
	return 1;
}

/**
 * Flags used by the lsatt command.
 *
 * \param \sa com_lsatt().
 */
enum lsatt_flags {
	/** Whether "-a" was given for the lsatt command. */
	LSATT_FLAG_SORT_BY_ID = 1,
	/** Whether "-l" was given for the lsatt command. */
	LSATT_FLAG_LONG = 2,
	/** Reverse sort order. */
	LSATT_FLAG_REVERSE = 4
};

/** Data passed to the action function of lsatt */
static int print_attribute(struct osl_table *table, struct osl_row *row,
		const char *name, void *data)
{
	struct afs_callback_arg *aca = data;
	unsigned flags = *(unsigned *)aca->query.data;
	struct osl_object bitnum_obj;
	int ret;

	if (!(flags & LSATT_FLAG_LONG)) {
		para_printf(&aca->pbout, "%s\n", name);
		return 1;
	}
	ret = osl(osl_get_object(table, row, ATTCOL_BITNUM, &bitnum_obj));
	if (ret < 0) {
		para_printf(&aca->pbout, "%s: %s\n", name, para_strerror(-ret));
		return ret;
	}
	para_printf(&aca->pbout, "%u\t%s\n", *(unsigned char*)bitnum_obj.data,
		name);
	return 1;
}

static int com_lsatt_callback(struct afs_callback_arg *aca)
{
	unsigned flags = *(unsigned *)aca->query.data;
	int ret;
	struct pattern_match_data pmd = {
		.table = attribute_table,
		.loop_col_num = ATTCOL_NAME,
		.match_col_num = ATTCOL_NAME,
		.patterns = {.data = (char *)aca->query.data + sizeof(flags),
			.size = aca->query.size - sizeof(flags)},
		.pm_flags = PM_NO_PATTERN_MATCHES_EVERYTHING,
		.data = aca,
		.action = print_attribute
	};
	if (flags & LSATT_FLAG_SORT_BY_ID)
		pmd.loop_col_num = ATTCOL_BITNUM;
	if (flags & LSATT_FLAG_REVERSE)
		pmd.pm_flags |= PM_REVERSE_LOOP;
	ret = for_each_matching_row(&pmd);
	if (ret < 0)
		goto out;
	if (pmd.num_matches == 0)
		ret = -E_NO_MATCH;
out:
	return ret;
}

int com_lsatt(struct command_context *cc)
{
	unsigned flags = 0;
	struct osl_object options = {.data = &flags, .size = sizeof(flags)};
	int i;

	for (i = 1; i < cc->argc; i++) {
		const char *arg = cc->argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strcmp(arg, "-i")) {
			flags |= LSATT_FLAG_SORT_BY_ID;
			continue;
		}
		if (!strcmp(arg, "-l")) {
			flags |= LSATT_FLAG_LONG;
			continue;
		}
		if (!strcmp(arg, "-r")) {
			flags |= LSATT_FLAG_REVERSE;
			continue;
		}
	}
	return send_option_arg_callback_request(&options, cc->argc - i, cc->argv + i,
		com_lsatt_callback, afs_cb_result_handler, cc);
}

struct addatt_event_data {
	const char *name;
	unsigned char bitnum;
};


static int com_addatt_callback(struct afs_callback_arg *aca)
{
	char *p;
	int ret = 1;
	size_t len;

	for (
		p = aca->query.data;
		p < (char *)aca->query.data + aca->query.size;
		p += len + 1
	) {
		struct osl_object objs[NUM_ATT_COLUMNS];
		struct osl_row *row;
		unsigned char bitnum;
		struct addatt_event_data aed;

		len = strlen(p);
		if (!len || p[len - 1] == '-' || p[len - 1] == '+') {
			para_printf(&aca->pbout, "invalid attribute name: %s\n", p);
			continue;
		}
		ret = get_attribute_bitnum_by_name(p, &bitnum);
		if (ret >= 0) {
			para_printf(&aca->pbout, "attribute \"%s\" already exists\n", p);
			continue;
		}
		if (ret != -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND)) /* error */
			goto out;
		objs[ATTCOL_BITNUM].size = 1;
		/* find smallest unused attribute */
		for (bitnum = 0; bitnum < 64; bitnum++) {
			objs[ATTCOL_BITNUM].data = &bitnum;
			ret = osl(osl_get_row(attribute_table, ATTCOL_BITNUM,
				&objs[ATTCOL_BITNUM], &row));
			if (ret == -OSL_ERRNO_TO_PARA_ERROR(E_OSL_RB_KEY_NOT_FOUND))
				break; /* this bitnum is unused, use it */
			if (ret < 0) /* error */
				goto out;
			/* this bit is already in use, try next bit */
		}
		if (bitnum == 64) {
			ret = -E_ATT_TABLE_FULL;
			goto out;
		}
		objs[ATTCOL_NAME].data = p;
		objs[ATTCOL_NAME].size = len + 1;
		ret = osl(osl_add_row(attribute_table, objs));
		if (ret < 0)
			goto out;
		aed.name = p;
		aed.bitnum = bitnum;
		ret = afs_event(ATTRIBUTE_ADD, &aca->pbout, &aed);
		if (ret < 0)
			goto out;
		greatest_att_bitnum = PARA_MAX(greatest_att_bitnum, (int)bitnum);
	}
out:
	if (ret < 0)
		para_printf(&aca->pbout, "%s: %s\n", p, para_strerror(-ret));
	return ret;
}

int com_addatt(struct command_context *cc)
{
	int ret;

	if (cc->argc < 2)
		return -E_ATTR_SYNTAX;
	ret = send_standard_callback_request(cc->argc - 1, cc->argv + 1,
		com_addatt_callback, afs_cb_result_handler, cc);
	if (ret < 0)
		send_strerror(cc, -ret);
	return ret;
}

static int com_mvatt_callback(struct afs_callback_arg *aca)
{
	char *old = aca->query.data;
	size_t size = strlen(old) + 1;
	char *new = old + size;
	struct osl_object obj = {.data = old, .size = size};
	struct osl_row *row;
	int ret;

	ret = osl(osl_get_row(attribute_table, ATTCOL_NAME, &obj, &row));
	if (ret < 0)
		goto out;
	obj.data = new;
	obj.size = strlen(new) + 1;
	ret = osl(osl_update_object(attribute_table, row, ATTCOL_NAME, &obj));
out:
	if (ret < 0)
		para_printf(&aca->pbout, "cannot rename %s to %s\n", old, new);
	else
		ret = afs_event(ATTRIBUTE_RENAME, &aca->pbout, NULL);
	return ret;
}

int com_mvatt(struct command_context *cc)
{
	if (cc->argc != 3)
		return -E_ATTR_SYNTAX;
	return send_standard_callback_request(cc->argc - 1, cc->argv + 1,
		com_mvatt_callback, afs_cb_result_handler, cc);
}

static int remove_attribute(struct osl_table *table, struct osl_row *row,
		const char *name, void *data)
{
	struct afs_callback_arg *aca = data;
	int ret;
	struct rmatt_event_data red = {.name = name};

	ret = get_attribute_bitnum_by_name(name, &red.bitnum);
	if (ret < 0) {
		para_printf(&aca->pbout, "cannot remove %s\n", name);
		return ret;
	}
	para_printf(&aca->pbout, "removing attribute %s\n", name);
	ret = osl(osl_del_row(table, row));
	if (ret < 0) {
		para_printf(&aca->pbout, "cannot remove %s\n", name);
		return ret;
	}
	return afs_event(ATTRIBUTE_REMOVE, &aca->pbout, &red);
}

static int com_rmatt_callback(struct afs_callback_arg *aca)
{
	int ret;
	struct pattern_match_data pmd = {
		.table = attribute_table,
		.patterns = aca->query,
		.loop_col_num = ATTCOL_BITNUM,
		.match_col_num = ATTCOL_NAME,
		.data = aca,
		.action = remove_attribute
	};
	ret = for_each_matching_row(&pmd);
	if (ret < 0)
		goto out;
	if (pmd.num_matches == 0)
		ret = -E_NO_MATCH;
out:
	return ret;
}

int com_rmatt(struct command_context *cc)
{
	if (cc->argc < 2)
		return -E_ATTR_SYNTAX;
	return send_standard_callback_request(cc->argc - 1, cc->argv + 1,
		com_rmatt_callback, afs_cb_result_handler, cc);
}

/**
 * Return a binary representation of the given attribute value.
 *
 * \param atts Pointer to the attribute value.
 * \param buf Result.
 *
 * This function prints a string of at most 64 characters plus the terminating
 * \p NULL character into \a buf which must be provided by the caller and at
 * least 65 bytes long. The "x" character is used for set attributes and "-" is
 * used for unset attributes.
 *
 * In practice, not all 64 attributes are defined. In this case, the function
 * only prints \a N + 1 characters where \a N is the greatest id of a defined
 * attribute.
 */
void get_attribute_bitmap(const uint64_t *atts, char *buf)
{
	int i;
	const uint64_t one = 1;

	for (i = 0; i <= greatest_att_bitnum; i++)
		buf[greatest_att_bitnum - i] = (*atts & (one << i))? 'x' : '-';
	buf[i] = '\0';
}

/**
 * Get a string containing the set attributes in text form.
 *
 * \param atts The attribute bitmap.
 * \param delim The delimiter to separate matching attribute names.
 * \param text Result pointer.
 *
 * \return Positive on success, negative on errors. If no attributes have
 * been defined, \a *text is NULL.
 */
int get_attribute_text(uint64_t *atts, const char *delim, char **text)
{
	int i, ret;
	const uint64_t one = 1;

	*text = NULL;
	if (greatest_att_bitnum < 0) { /* no attributes available */
		*text = para_strdup("(no attributes available)");
		return 1;
	}
	for (i = 0; i <= greatest_att_bitnum; i++) {
		unsigned char bn = i;
		struct osl_object obj = {.data = &bn, .size = 1};
		struct osl_row *row;

		if (!(*atts & (one << i)))
			continue;
		ret = osl(osl_get_row(attribute_table, ATTCOL_BITNUM, &obj, &row));
		if (ret < 0)
			goto err;
		ret = osl(osl_get_object(attribute_table, row, ATTCOL_NAME, &obj));
		if (ret < 0)
			goto err;
		if (*text) {
			char *tmp = make_message("%s%s%s", *text, delim, (char *)obj.data);
			free(*text);
			*text = tmp;
		} else
			*text = para_strdup(obj.data);
	}
	if (!*text) /* no attributes set */
		*text = para_strdup("");
	return 1;
err:
	free(*text);
	return ret;
}

static int att_logical_or(struct osl_row *row, void *data)
{
	uint64_t *att_mask = data, one = 1;
	struct osl_object bitnum_obj;
	int ret = osl_get_object(attribute_table, row, ATTCOL_BITNUM, &bitnum_obj);

	if (ret < 0)
		return ret;
	*att_mask |= one << *(unsigned char *)bitnum_obj.data;
	return 0;
}

/**
 * Compute the attribute bit mask and check each afs info bitmap.
 *
 * \param aca The query field of \a aca is ignored.
 *
 * This iterates over all attributes in the attribute table and computes the
 * logical or of 1 << b where b is the bit number of the attribute. The
 * resulting bit mask is passed to aft_check_attributes() which performs the
 * actual check.
 *
 * \return Standard.
 *
 * \sa \ref aft_check_attributes().
 */
int attribute_check_callback(struct afs_callback_arg *aca)
{
	int ret;
	uint64_t att_mask = 0; /* bits corresponding to a attributes */

	ret = osl_rbtree_loop(attribute_table, ATTCOL_BITNUM, &att_mask,
		att_logical_or);
	if (ret < 0) {
		PARA_ERROR_LOG("attribute table loop failed: %s\n",
			para_strerror(-ret));
		return ret;
	}
	return aft_check_attributes(att_mask, &aca->pbout);
}

/**
 * Close the attribute table.
 *
 * \sa osl_close_table().
 */
static void attribute_close(void)
{
	osl_close_table(attribute_table, OSL_MARK_CLEAN);
	attribute_table = NULL;
}

/**
 * Open the attribute table.
 *
 * \param dir The database directory.
 *
 * \return Positive on success, negative on errors.
 *
 * \sa osl_open_table().
 */
static int attribute_open(const char *dir)
{
	int ret;

	attribute_table_desc.dir = dir;
	ret = osl(osl_open_table(&attribute_table_desc, &attribute_table));
	greatest_att_bitnum = -1; /* no atts available */
	if (ret >= 0) {
		find_greatest_att_bitnum();
		return ret;
	}
	attribute_table = NULL;
	if (ret == -OSL_ERRNO_TO_PARA_ERROR(E_OSL_NOENT))
		return 1;
	return ret;
}

static int attribute_create(const char *dir)
{
	attribute_table_desc.dir = dir;
	return osl(osl_create_table(&attribute_table_desc));
}

/**
 * Initialize the attribute table structure.
 *
 * \param t The table structure to initialize.
 */
void attribute_init(struct afs_table *t)
{
	t->open = attribute_open;
	t->close = attribute_close;
	t->create = attribute_create;
}
